@@ -1,6 +1,6 @@
 package Auto::Mata;
 # ABSTRACT: A simple, reliable finite state machine
-$Auto::Mata::VERSION = '0.01';
+$Auto::Mata::VERSION = '0.02';
 
 use v5.10;
 use strict;
@@ -31,8 +31,13 @@ my $State = declare 'State', as Tuple[$Ident, Any];
 my $Type  = declare 'Type',  as InstanceOf['Type::Tiny'];
 my $Code  = declare 'Code',  as CodeRef;
 coerce $Code, from Undef, via { sub { $_ } };
+coerce $Type, from Undef, via { Any };
+
+my $Transition = declare 'Transition', as Dict[
+  initial   => $Type,
+  transform => $Code,
+];
 
-my $Transition = declare 'Transition', as Tuple[$Type, $Type, $Code];
 my $Automata = declare 'Automata', as Dict[
   ready => Maybe[$Ident],
   term  => Maybe[$Ident],
@@ -55,6 +60,8 @@ sub machine (&) {
     validate();
   };
 
+  my $terminal = declare 'Terminal', as Tuple[Enum[$fsm{term}], Any];
+
   #-----------------------------------------------------------------------------
   # Build the transition engine
   #-----------------------------------------------------------------------------
@@ -64,20 +71,14 @@ sub machine (&) {
     # Create type constraints for each "from" state to validate that the
     # machine's state is consistent after each transition.
     #---------------------------------------------------------------------------
-    my @next_types;
-    foreach my $to (keys %{$map{$from}}) {         # $to = each state $from can transition to
-      foreach my $next (keys %{$map{$to}}) {       # $next = each state $to can transition to
-        if ($next eq $fsm{term}) {                 # Termination matches any result since there
-          push @next_types, Any;                   #   are no further states to validate against
-        } else {
-          push @next_types, $map{$to}{$next}->[0]; # Initial state constraints for each $next
-        }
-      }
+    my @next;
+    foreach my $to (keys %{$map{$from}}) {
+      push @next, $to eq $fsm{term}
+        ? $terminal
+        : map { $map{$to}{$_}{initial} } keys %{$map{$to}};
     }
 
-    my $next = declare "Next_State_After_$from", as @next_types
-      ? reduce { $a | $b } @next_types
-      : Any;
+    my $next = declare "${from}_FINAL", as reduce { $a | $b } @next;
 
     #---------------------------------------------------------------------------
     # Create a type constraint that matches each possible initial "from" state.
@@ -85,27 +86,23 @@ sub machine (&) {
     # for that transisiton.
     #---------------------------------------------------------------------------
     foreach my $to (keys %{$map{$from}}) {
-      my ($match, $on, $with) = @{$map{$from}{$to}};
+      my $match = $map{$from}{$to}{initial};
+      my $with  = $map{$from}{$to}{transform};
 
       push @match, $match, sub {
-        debug("%s -> %s: %s", $_->[0], $to, explain($_->[1]));
         my ($from, $input) = @$_;
+        debug('%s -> %s: %s', $from, $to, explain($input));
 
         do { local $_ = $input; $input = $with->() };
-
         my $state = [$to, $input];
 
         if (defined(my $error = $next->validate($state))) {
-          my @msg = (
-            sprintf('Transition from %s to %s resulted in an invalid state.', $from, $to),
-            sprintf('Current state is: %s', explain($state)),
-            sprintf('Type constraint returned an error: %s', $error),
-          );
+          debug(" -$_") foreach @{$next->validate_explain($state, 'FINAL_STATE')};
 
-          push(@msg, join "\n", map { " -$_" } @{$next->validate_explain($_, 'NEXT STATE')})
-            if $DEBUG;
-
-          croak join("\n", @msg);
+          croak join "\n",
+            sprintf('Transition from %s to %s produced an invalid state.', $from, $to),
+            sprintf('Attempted to move from %s to %s', explain($_), explain($state)),
+            sprintf($error);
         }
 
         return @$state;
@@ -150,11 +147,15 @@ sub transition ($%) {
   croak "transition from state $from to $to is already defined"
     if exists $_->{map}{$from}{$to};
 
-  my $name  = sprintf('%s_to_%s', $from, $to);
-  my $match = declare $name, as Tuple[Enum[$from], $on];
+  my $state   = declare "STATE_${from}", as Enum[$from];
+  my $initial = declare "${from}_TO_${to}_INITIAL", as Tuple[$state, $on];
 
   $_->{map}{$from} //= {};
-  $_->{map}{$from}{$to} = [$match, $on, $with];
+
+  $_->{map}{$from}{$to} = {
+    initial   => $initial,
+    transform => $with,
+  };
 }
 
 #-------------------------------------------------------------------------------
@@ -165,8 +166,8 @@ sub assert_in_the_machine {
   croak 'cannot be called outside a state machine definition block' unless $_;
 
   unless (!defined(my $msg = $Automata->validate_explain($_, '$_'))) {
-    debug('Invalid machine state detected: %s', join("\n", map {" -$_"} @$msg)) if $msg;
-    croak 'Invalid machine definition';
+    debug('Invalid machine state detected: %s', join("\n", map {" -$_"} @$msg));
+    croak 'invalid machine definition';
   }
 }
 
@@ -248,7 +249,7 @@ Auto::Mata - A simple, reliable finite state machine
 
 =head1 VERSION
 
-version 0.01
+version 0.02
 
 =head1 SYNOPSIS
 
